import tkinter as tk
from typing import Optional

from bredon.controller import *


tk.Canvas.create_circle = lambda self, x, y, r, **kwargs: self.create_oval(x - r, y - r, x + r, y + r, **kwargs)


class ViewSquare:
    def __init__(self, master, i, j, width=SQUARE_SIZE, height=SQUARE_SIZE):
        self.width = width
        self.height = height
        self.master = master
        self.i, self.j = i, j
        self.ix, self.jy = self.i * SQUARE_SIZE + 2, self.j * SQUARE_SIZE + 2
        self.tags = str(i) + "." + str(j)
        self.ids = []
        self.nridx = None

    def render_tile(self, tile, idx, offset_x=0.0, offset_y=0.0):
        if tile is None: return
        o = OFFSET_STEP * idx

        x1, y1 = offset_x + self.ix + S - N, offset_y + self.jy + S - N - o + D
        x2, y2 = offset_x + self.ix + S + N, offset_y + self.jy + S + N - o + D
        if tile.stone == FLAT:
            return self.master.canvas.create_rectangle(x1, y1, x2, y2, outline=tile.color.flip().name, fill=tile.color.name)
        elif tile.stone == STAND:
            if tile.color == Colors.WHITE:
                points = (
                    x1 + N * 1.5, y1,
                    x2, y1 + N * 0.5,
                    x1 + N * 0.5, y2,
                    x1, y1 + N * 1.5
                )
            else:
                points = (
                    x1 + N * 0.5, y1,
                    x2, y1 + N * 1.5,
                    x1 + N * 1.5, y2,
                    x1, y1 + N * 0.5
                )

            return self.master.canvas.create_polygon(*points, fill=tile.color.name, outline=tile.color.flip().name,
                                                     tags=(self.tags, idx))
        else:
            return self.create_circle(*self.find_center(idx, offset_x, offset_y), S / 2, fill=tile.color.name,
                                      outline=tile.color.flip().name, tags=(self.tags, idx))

    def find_center(self, idx, offset_x=0.0, offset_y=0.0):
        o = OFFSET_STEP * idx
        return offset_x + self.ix + S, offset_y + self.jy + S - o + D

    def render(self, active=False, possible=False):
        self.rect = self.master.canvas.create_rectangle(self.ix, self.jy, self.ix + SQUARE_SIZE, self.jy + SQUARE_SIZE,
                                                        fill="green" if active else "white")
        if possible:
            self.circle = self.create_circle(self.ix + S, self.jy + S, S - 5, outline="blue", width=5)
        self.ids = []
        tiles = self.get_tiles(self.master.board)[:]
        if self.nridx is not None:
            tiles.insert(self.nridx, None)
        self.ids = [self.render_tile(tile, idx) for idx, tile in enumerate(self.get_tiles(self.master.board), start=1)]

    def create_circle(self, x, y, r, **kwargs):
        return self.master.canvas.create_circle(x, y, r, **kwargs)

    def get_square(self, board):
        return board.board[self.i][self.j]

    def get_tiles(self, board):
        return self.get_square(board).tiles

    def S(self, b):
        return self.tags + " " + str(self.get_tiles(b))


class ViewBoard(tk.Frame):
    def __init__(self, master):
        self.board = master.board
        self.size = master.board.size
        self.squares = []

        tk.Frame.__init__(self, master, width=(self.size + 2) * SQUARE_SIZE, height=(self.size + 2) * SQUARE_SIZE)
        self.pack_propagate(0)
        self.canvas = tk.Canvas(self, width=self.size * SQUARE_SIZE, height=self.size * SQUARE_SIZE)
        # self.canvas.pack()
        self._init_gui()
        self.actives = [False for _ in range(self.size ** 2)]
        self.move = None

        # Saved states for movement
        self.i = 1
        self.grabbed: Optional[bool, ViewSquare] = False
        self.grabbed_first = False
        self.direction = False
        self.nridx = 0
        self.moves = []

        # Canvas bindings
        self.canvas.bind("<2>", self.master.exec)
        self.canvas.bind("<1>", self.click)
        self.canvas.bind("<Escape>", self.clear)

    def _init_gui(self):
        self.input = tk.Entry(self, font = "Courier 12")  #, width = SQUARE_SIZE * 2 // 12)
        self.input.bind("<Return>", self.master.exec)
        self.exec_button = tk.Button(self, text="Exec", font="Courier 12", command=self.master.exec, width=7)
        self.clear_button = tk.Button(self, text="Clear", font="Courier 12", command=self.clear, width=7)

        self.canvas.grid(row=1, column=1, columnspan=self.size, rowspan=self.size)

        self.input.grid(row=6, column=1, columnspan=2)
        self.exec_button.grid(row=6, column=4)
        self.clear_button.grid(row=6, column=5)

        spaces = " "*5
        self.row_labels = tk.Label(self, text=spaces+(spaces*2).join(ascii_lowercase[:self.size])+spaces,
                                   font="Courier 12", height=1).grid(column=1, row=0, columnspan=self.size)

        self.col_labels = [tk.Label(self, text=i + 1, font="Courier 12", height=1, width=1)  #, bd=5, relief=tk.GROOVE)
                           # .grid(column=0, row=i + 1)
                           for i in range(self.size)]
        for i, l in enumerate(self.col_labels, start=1):
            l.place_configure(x=i*SQUARE_SIZE)
            l.grid(column=0, row=i)


        for i in range(self.size):
            for j in range(self.size):
                self.squares.append(ViewSquare(self, i, j))

        self.canvas.focus_set()

    def click(self, event):
        def _run(S, f=False):
            self.input.delete(0, tk.END)
            self.input.insert(0, S)
            self.move = None
            self.render()  #flip_color(self.master.get_color()))
            self.b = self.board.copy()
            self.board = self.master.board.copy()
            self.board.force_str(self.input.get(), self.master.get_color())
            self.execute(self.input.get().strip(), self.master.players[self.master.player], self.b)
            self.render()  #flip_color(self.master.get_color()))
            self.move = None
            if not f: self.board = self.b

        stones = self.master.stones[self.master.get_color()]
        self.i = (self.i + 1) % len(stones)
        x, y = event.x // SQUARE_SIZE, event.y // SQUARE_SIZE
        sq = self.squares[x * self.size + y]
        t = sq.get_tiles(self.board)
        if not self.grabbed:
            if len(t) == 0:
                _run(stones[self.i - 1] + ascii_lowercase[x] + str(y + 1))
            elif t[-1].color == self.master.players[self.master.player].color:
                self.grabbed = sq
                self.grabbed_first = self.grabbed
                self.grabbed.nridx = self.nridx = 1
                self.render()  #flip_color(self.master.players[self.master.player].color))
        else:
            x1, y1 = sq.i, sq.j
            x, y = self.grabbed.i, self.grabbed.j
            a, b = abs(x-x1), abs(y-y1)
            if (a <= 1 and b <= 1) and ((a == 1) ^ (b == 1)):
                if a == 0:
                    direction = " -+"[y-y1]
                else:
                    direction = " <>"[x-x1]
                if not self.direction:
                    self.direction = direction
                elif self.direction != direction:
                    return
                x, y = self.grabbed_first.i, self.grabbed_first.j
                self.moves.append(1)
                S = str(self.nridx) + ascii_lowercase[x] + str(y + 1) + direction
                if self.moves:
                    S += ''.join(map(str, self.moves))
                _run(S, f=True)

                if self.grabbed.nridx > 1:
                    self.grabbed.nridx -= 1
                    self.switch_g()
                    self.render()  #flip_color(self.master.players[self.master.player].color))
            elif a == 0 and b == 0:
                if not self.direction and len(self.grabbed.get_tiles(self.board)) > self.grabbed.nridx:
                    self.grabbed.nridx += 1
                    self.nridx += 1
                elif sum(self.moves) < len(self.grabbed_first.get_tiles(self.board)):
                    self.moves[-1] += 1
                self.render()  #flip_color(self.master.players[self.master.player].color))

    def clear(self, r=True):
        if not isinstance(self.grabbed, bool):
            self.grabbed.nridx = self.nridx = 0
        self.i = 1
        self.grabbed: Optional[bool, ViewSquare] = False
        self.grabbed_first = self.direction = False
        self.board = self.master.board
        self.moves = []
        if r:
            self.render()  #flip_color(self.master.players[self.master.player].color))

    def switch_g(self):
        N = self.grabbed.nridx
        x, y = self.grabbed.get_square(self.board) \
            .next(self.direction, self.size)
        self.grabbed = self.squares[y * self.size + x]
        self.grabbed.nridx = N

    def possibles(self, color):
        if not bool(self.grabbed):
            for sq in self.squares:
                t = sq.get_tiles(self.board)
                if t:
                    yield t[-1].color == color
                else:
                    yield True
        else:
            sq = self.grabbed.get_square(self.board)
            ps = [False] * (self.size ** 2)
            for D in DIRS:
                x, y = sq.next(D, self.size)
                t = self.board.board[x][y].tiles
                if t:
                    ps[y * self.size + x] = t[-1].stone is FLAT
                else:
                    ps[y * self.size + x] = True

            yield from ps

    def execute(self, move, player, old_board):
        new_board = self.board.copy()
        new_board.force_str(move, player)
        idx = 0
        for rn, ro in zip(new_board.board, old_board.board):
            for nsq, sq in zip(rn, ro):
                if str(nsq) != str(sq):
                    self.actives[idx] = True
                idx += 1
        # self.render()
        self.move = move, old_board

    def animate(self, move, old_board):
        m = str_to_move(move)
        sq = int(ascii_lowercase.index(m.col)), int(m.row) - 1
        moves = m.moves if m.moves else [m.total]
        t = m.total
        for i, move in enumerate(moves):
            stop = -(t - move)
            if stop == 0:
                stop = None
            self._animate(sq, m.direction, i, move, old_board, slice(-t, stop))
            t -= move

    def _animate(self, sq, direction, i, N, old_board, idxs):
        vis = self.get_square(sq)
        tiles = vis.get_tiles(old_board)[idxs]
        ids = vis.ids[idxs]

        nvis = Square(*sq).tru_next(direction, self.size)
        for _ in range(i):
            nvis = Square(*nvis).tru_next(direction, self.size)
        nts = self.get_square(nvis).get_tiles(old_board)

        step = SQUARE_SIZE / ANIM_STEPS * (i + 1)
        xi, yi = None, None
        if direction == UP:
            step -= (3 * len(nts)) / ANIM_STEPS
            xi, yi = step, 0
        elif direction == DOWN:
            xi, yi = -step, 0
        elif direction == RIGHT:
            xi, yi = 0, -step
        elif direction == LEFT:
            xi, yi = 0, step

        for k in range(2, ANIM_STEPS):
            for _id in ids:
                self.canvas.delete(_id)
            ids = [vis.render_tile(tile, _idx, -(yi * k), (xi * k) - ((_idx - 1) * 3) * (k / ANIM_STEPS))
                   for _idx, tile in enumerate(tiles, start=1)]
            self.update()

    def get_square(self, sq) -> ViewSquare:
        return self.squares[sq[0] * self.size + sq[1]]

    def render(self):  #, color):
        color = self.master.get_color()
        if self.move is not None:
            if str_to_move(self.move[0]).direction is not None:
                self.animate(self.move[0], self.move[1])
        self.canvas.delete("all")

        for S, a, p in zip(self.squares, self.actives, self.possibles(color)):
            S.render(a, p)
        self.canvas.create_line(3, 3, self.winfo_width(), 3, tags="line")
        self.canvas.create_line(3, 3, 3, self.winfo_height(), tags="line")
        self.actives = [False for _ in range(self.size ** 2)]


class _Canvas(tk.Canvas):
    def __init__(self, master, width, height, bd=5, relief=tk.GROOVE):
        self.master = master
        self.board = master.board
        self.players = master.players
        self.width = width
        self.height = height
        self.name1, self.name2 = self.get_name(0), self.get_name(1)

        tk.Canvas.__init__(self, master, bd=bd, relief=relief, width=width, height=height)

    def get_player(self, player):
        return self.players[player]

    def get_name(self, player):
        return self.get_player(player).name.title()

    def calc_stones(self, player):
        p = self.get_player(player)
        return self.master.board.caps - p.caps, self.master.board.stones - p.stones

    def render(self):
        raise NotImplementedError()


class FlatCanvas(_Canvas):
    def __init__(self, master):
        _Canvas.__init__(self, master, master.board.size * SQUARE_SIZE, 30, bd=0)

    def render(self):
        self.delete("all")
        w, b = self.board.count_flats(Colors.WHITE), self.board.count_flats(Colors.BLACK)
        try:
            S = (w / (w + b)) * self.width
        except ZeroDivisionError:
            S = self.width / 2
        if S != 0:
            self.create_rectangle(0, 0, S, self.height, fill=Colors.WHITE.name, outline=Colors.BLACK.name)
            self.create_text(S/2, self.height / 2, text=w)
        if S != self.width:
            self.create_rectangle(S, 0, self.width, self.height, fill=Colors.BLACK.name)
            self.create_text(S + (self.width-S)/2, self.height / 2, text=b, fill=Colors.WHITE.name)


class TilesCanvas(_Canvas):
    def __init__(self, master, color):
        self.step = TILE_SIZE / 4
        _Canvas.__init__(self, master,
                         width=len(master.players[0].name)*12+10,
                         height=SQUARE_SIZE * master.board.size)
        self.player = [i for i, p in enumerate(master.players) if p.color == color][0]

    def render(self):
        self.delete("all")
        self.create_text(self.get_x(self.player), 40, font=("Ubuntu Mono", 24),
                         text="{}\n{}+{}".format(self.get_name(self.player), *self.calc_stones(self.player)))
        self.draw_tiles(self.player)

    def get_x(self, player):
        return TILE_SIZE

    def draw_tiles(self, player):
        p = (player) % 2
        x2, y2 = self.get_x(p) + 20, self.height
        x1, y1 = x2 - TILE_SIZE, y2 - TILE_SIZE
        S = TILE_SIZE / 2
        for _ in range(self.calc_stones(p)[1]):
            self.create_rectangle(x1, y1, x2, y2, fill=Colors(player).name, outline=Colors(p).flip().name)
            y1 -= self.step
            y2 -= self.step
        for _ in range(self.calc_stones(p)[0]):
            self.create_circle(x2-S, y2-S, S, fill=Colors(player).name, outline=Colors(p).flip().name)
            y2 -= self.step


class PtnCanvas(_Canvas):
    def __init__(self, master):
        _Canvas.__init__(self, master, width=150, height=SQUARE_SIZE * master.board.size)

    def render(self):
        self.delete("all")
        self.create_text(12, 12, anchor=tk.NW, font=("Courier", 12),
                         text=str(self.master.ptn))

class ViewGame(tk.Tk, Game):
    def __init__(self, **kw):
        tk.Tk.__init__(self)
        Game.__init__(self, **kw)
        self._init_gui()

        self.viz()
       
    def _init_gui(self, **kw):
        for key, value in kw.items():
            self.__setattr__(key, value)
        self.wm_title("tak")
        
        self.vboard = ViewBoard(self)
        self.btiles = TilesCanvas(self, Colors.BLACK)
        self.wtiles = TilesCanvas(self, Colors.WHITE)
        self.flats = FlatCanvas(self)
        self.vptn = PtnCanvas(self)

        self.flats.grid(row=0, column=1, columnspan=self.board.size)
        self.vboard.grid(row=2, column=1)
        self.btiles.grid(row=2, column=0, rowspan=self.board.size)
        self.wtiles.grid(row=2, column=self.board.size+1, rowspan=self.board.size)
        self.vptn.grid(row=2, column=self.board.size+2, rowspan=self.board.size)

        self.turn = self.player = 0
        self.running = False

        self.first = True
        self.event = None

        self.run = self.mainloop

    def _clear_gui(self):
        for canvas in [self.flats, self.btiles, self.vboard, self.wtiles, self.vptn]:
            canvas.grid_forget()
        del self.vboard
        del self.btiles
        del self.wtiles
        del self.flats
        del self.vptn

    def exec(self, *event, txt=None, is_ai=False):
        if not txt:
            txt = self.vboard.input.get()
        if not is_ai and not txt:
            return
        if not self.running:
            return
        elif self.player == 0:
            self.turn += 1
        if "TPS" in txt:
            self._clear_gui()
            self._init_gui(**self.exec_tps(*parse_tps(txt)))
            self.vboard.clear()
            self.running = True
        else:
            p = self.players[self.player]
            try:
                if not self._run(p, self.turn, input_fn=lambda _: txt):
                    return
            except Exception as e1:
                try:
                    self.board = Board.from_moves(exec_road(txt, self.board.size), self.board.size,
                                                  colors=repeat(self.get_color()),
                                                  board=self.board)
                    self.vboard.board = self.board
                except Exception as e2:
                    print("Unexpected error while parsing")
                    print(e1)
                    print("While parsing")
                    print(e2)
            print(self.ptn)
            self.viz()
        self.vboard.input.delete(0, "end")
        self.vboard.i = 0
        if self.vboard.grabbed:
            self.vboard.clear(r=False)
        print(self.board)
        self.vboard.board = self.board
        self.viz()

        w = self.board.winner(self.players, t=True)
        if w:
            print(w, "won!")

        if "TPS" not in txt:
            self.player = (self.player + 1) % 2
            if isinstance(self.players[self.player], BaseAI):
                self.exec(is_ai=True)

    def viz(self):
        for canvas in [self.flats, self.btiles, self.vboard, self.wtiles, self.vptn]:
            canvas.render()

        self.update_idletasks()
        self.update()

    def mainloop(self, n=0):
        self.running = True
        if isinstance(self.players[self.player], BaseAI):
            self.exec(is_ai=True)
        super(tk.Tk, self).mainloop(n=n)

    def get_color(self):
        c = self.players[self.player].color
        if self.turn < 2 and self.turn == self.player:
            c = c.flip()
        return c



import enum


# Colors
class Colors(enum.Enum):
    WHITE = 0
    BLACK = 1

    def flip(self):
        return Colors(1 - self.value)

    def __str__(self):
        return self.name  #[0]

    @staticmethod
    def of(s):
        return Colors[s]


# PTN
MARKS = '?!'

# Moves
DIRS = '+-<>'
UP, DOWN, LEFT, RIGHT = DIRS
STONES = 'FCS'
FLAT, CAP, STAND = STONES
EMPTY = ' '

# Graphics
TILE_SIZE   = 40
SQUARE_SIZE = 75
ANIM_STEPS  = 30
OFFSET_STEP = 3
PAD_STEP    = 5

# What a great coder
N, S, D = TILE_SIZE / 2, SQUARE_SIZE / 2, PAD_STEP

# AI
MAX_N = 1234567890
THRESHOLD = 10000

# Controller
HUMAN = "human", None
AI = lambda n: ("AI", n)

# Misc
SIZES = {
    3: [10, 0],
    4: [15, 0],
    5: [21, 1],
    6: [30, 1],
    7: [40, 2],
    8: [50, 2]
}



import pickle
from model import *
import numpy as np

def gen_classes(size):
    for c in ascii_lowercase[:size]:
        for r in range(1, size + 1):
            for s in STONES:
                yield s + c + str(r)
            for t in range(1, size + 1):
                for ms in sums(t):
                    for d in DIRS:
                        yield str(t) + c + str(r) + d + ''.join(map(str, ms))

classes = list(gen_classes(4))

def board_to_vector(ptn):
    b = np.zeroes(len(classes))
    for move in ptn:
        b[classes.index(move)] += 1 
    return b

def move_to_vector(m):
    v = np.zeroes(len(classes))
    v[classes.index(m)] = 1
    return v

def load_features(fn):
    print("Reading:", fn)
    boards, moves = [], []
    ptn = PTN()
    for m in load_moves_from_file(fn):
        ptn.append(m)
        boards.append(board_to_vector(ptn))
        moves.append(move_to_vector(m))
    return boards, moves


def handle_features(fn):
    boards, moves = load_features(fn)
    moves = moves[1:]
    boards = boards[:-1]
    return boards, moves

def create_feature_sets_and_labels(files, test_size = 0.1):
    features = []
    for file in files:
        features.extend(zip(*handle_features(file)))
    random.shuffle(features)
    features = np.array(features)

    testing_size = int(test_size*len(features))

    train_x = list(features[:,0][:-testing_size])
    train_y = list(features[:,1][:-testing_size])
    test_x = list(features[:,0][-testing_size:])
    test_y = list(features[:,1][-testing_size:])

    return train_x,train_y,test_x,test_y

if __name__ == '__main__':
    files = []
    train_x, train_y, test_x, test_y = create_feature_sets_and_labels(files)
    with open('note_features.pickle','wb') as f:
        pickle.dump([train_x, train_y, test_x, test_y],f)



from bredon.model import *
import time
from itertools import cycle

class Game:
    def __init__(self, size=5, board=None, white=HUMAN, black=AI(3)):
        self.board = Board(size) if board is None else board
        self.player_1: Player = self._init_player(Colors.WHITE, white)
        self.player_2: Player = self._init_player(Colors.BLACK, black)
        self.players = [self.player_1, self.player_2]

        self.stones = {i: FLAT for i in Colors}
        self.ptn = PTN()
        self.player_order = cycle(self.players)
        self.args = white, black

    def _init_player(self, color, types) -> Player:
        name, depth = types
        if name == HUMAN[0]:
            return Player(self.board, color)
        else:
            return MinimaxAI(self.board, color, depth)

    def viz(self):
        return

    def _run(self, player, turn, input_fn=input):
        self.viz()
        t = time.time()
        if turn <= 1:
            m, c = player.pick_opposing_move(turn, input_fn=input_fn)
            self.stones[c] = STONES
        else:
            m = player.pick_move(turn, input_fn=input_fn)
            c = player.color
            if CAP in str(m):
                self.stones[c] = FLAT + STAND
        print(time.time() - t)
        player._do(m, c)
        self.ptn.append(m)
        return m

    def run(self):
        self.ptn.clear()
        turn = 1
        # while True:
            # for player in self.players:
        for player in self.player_order:
            print(self.ptn)
            self._run(player, turn)
            w = self.board.winner(self.players, t=True)
            if w:
                print(self.ptn)
                self.viz()
                print(w, "won!")
                return
            turn += 0.5

    def exec_tps(self, board, move, turn, w_stones, b_stones):
        self.ptn = PTN(turn=int(turn))
        self.turn = int(turn)
        if int(move) == 2:
            def _p_o():
                yield self.player_2
                yield from chain(self.players)
            self.player_order = _p_o()
        else:
            self.player_order = chain(self.player_order)
        self.player = int(move) - 1
        self.player_1.board, self.player_2.board = board, board
        self.player_1.stones, self.player_1.caps = w_stones
        self.player_2.stones, self.player_2.caps = b_stones
        # self.viz()
        return dict(size=len(board.board[0]), board=board, white=self.args[0], black=self.args[1])


class TextGame(Game):
    def viz(self):
        print(self.board)
        print(self.board.generate_tps())



from .env import *
from ..model import *
from random import sample
from tqdm import *

class TakAgent(Player):
    def __init__(self, **kwargs):
        self.env = kwargs.pop("env")
        Player.__init__(self, **kwargs)

class HypoTakAgent(TakAgent):
    def __init__(self, **kwargs):
        # assert 1 == 'https://github.com/AxeldeRomblay/MLBox'
        self.depth = kwargs.pop("depth")
        self.env = kwargs.pop("env")
        self.other = None
        TakAgent.__init__(self, **kwargs)

    def flip_player(self, p):
        if p is self:
            return self.other
        return self

    def solve(self):
        if not self.other:
            self.other = [p for p in self.env.players if p is not self][0]
        # print("Solving!")
        actions = list(self.env.get_actions(self))
        self.rewards = {str(i): 0 for i in actions}
        hypo = hypoenv(self.env)
        for a in tqdm(actions):
            # print(a, end=': ')
            hypo.hypostep(a)
            self._solve(self.depth - 1, hypoenv(hypo), str(a), self.other) #, float("-inf"), float("inf"))
            # print(self.rewards[str(a)])
            hypo.unstep()
        return self.get_best_play()

    def get_best_play(self):
        # r = {j:i for i, j in self.rewards.items()}
        r = {}
        for i, j in self.rewards.items():
            r.setdefault(j, []).append(i)
        return r[max(r)]

    def _solve(self, depth, env, super_move, player): #, alpha, beta):
        # env.render()
        # input()
        if depth == 0: return
        actions = env.get_actions(player)
        n, c = 0, 0
        neg = player is not self
        for a in actions:
            try:
                s, r, d, i = env.hypostep(a)
                if r != 0:
                    # if neg:
                    #     r = -r
                    #     if r < beta:
                    #         beta = r
                    # elif r > alpha:
                    #     alpha = r
                    # if beta <= alpha:
                    #     # print("pruned!")
                    #     # if c != 0: self.rewards[super_move] += n / c
                    #     # return
                    #     break
                    n += r
                    c += 1
                    self._solve(depth - 1, hypoenv(env), super_move, self.flip_player(player)) #, alpha, beta)
                env.unstep()
            except TypeError as te: pass # invalid move
        if c != 0:
            self.rewards[super_move] += n / c

def make_agent(env, color, class_=TakAgent, **kwargs):
    return class_(env=env, color=color, board=env.board, **kwargs)

def main():
    env = TakEnv()
    env.add_player(make_agent(env, Colors.WHITE, depth=3))
    env.add_player(make_agent(env, Colors.BLACK, depth=3))

    env.reset()

    p1, p2 = env.players

    # env.step(Action(str_to_move("a1"), p1))
    # env.step(Action(str_to_move("b3"), p2))
    env.step(Action(str_to_move("a1"), p1))
    env.step(Action(str_to_move("a2"), p1))
    env.step(Action(str_to_move("a3"), p1))
    env.step(Action(str_to_move("a4"), p1))
    env.step(Action(str_to_move("b1"), p2))
    env.step(Action(str_to_move("c1"), p2))
    env.step(Action(str_to_move("d1"), p2))
    a = p2.solve()
    print(a)
    print(p2.rewards)
    '''
    while not env.done:
        env.render()
        env.step(Action.of(random.choice(p1.solve())))
        env.render()
        env.step(Action.of(random.choice(p2.solve())))
    '''



import unittest
from model import *
from itertools import repeat


class TestAI(unittest.TestCase):
    def test_block(self):
        b = Board.from_moves(exec_road("a", 4, False)[:-1], 4)
        ai = MinimaxAI(b, Colors.BLACK)
        self.assertEqual(str(ai.pick_move(4)), "a4", "Error: AI did not block road threat A")

    def test_take(self):
        b = Board.from_moves(exec_road("a", 4, False)[:-1], 4)
        ai = MinimaxAI(b, Colors.WHITE)
        self.assertEqual(str(ai.pick_move(4)), "a4", "Error: AI did not take road threat A")



if __name__ == '__main__':
    unittest.main()



from .env import *
from .actions import *
from .agent import *

from gym.envs.registration import register

for i in range(3, 9):
    print("registering", i)
    register(
        id=f'Tak{i}x{i}-v0',
        entry_point='bredon.env:TakEnv',
        kwargs={
            'size': i,
        }
    )


import os

URL = "playtak.com/games/{}.ptn"
DIR = "/Users/chervjay/Documents/GitHub/Bredon/neural/games/"
CMD = "wget -P {} {}"
def get(i):
    if os.path.exists(DIR + str(i) + ".ptn"):
        print("File exists")
        return
    run = CMD.format(DIR, URL).format(i)
    print("Executing", run)
    os.system(run)


for i in range(1, 100000):
    get(i)



from .notation import *
from .player   import *
from .utils    import *

def load_moves_from_file(filename, out=False):
    with open(filename) as file:
        ptn_str = file.read()
        _, s = ptn_str.split("Size")
        ptn_str = s.split("\n")
        size = int(s[2])
        b = Board(size)
        curr_player = Colors.WHITE
        ptn = PTN()
        for iturn, turn in enumerate(ptn_str[2:]):
            if 'R' in turn:
                break
            for imove, move in enumerate(turn.split(" ")[1:]):  # Exclude the round number
                if move:
                    if iturn == 0:
                        curr_player = [Colors.BLACK, Colors.WHITE][imove]
                    elif iturn == 1 and imove == 0:
                        curr_player = Colors.WHITE

                    b.force(b.parse_move(str_to_move(move), curr_player))

                    if out:
                        print(move)
                        print(str_to_move(move))
                        print(b)
                        print("Road?:", b.road())

                    if curr_player == Colors.BLACK:
                        curr_player = Colors.WHITE
                    else:
                        curr_player = Colors.BLACK
                    ptn.append(move)
                    yield b, move
    yield b, ptn

def exec_road(threat, size, return_size=False):
    """
    Threat types:
    1) column or row (straight) e.g. "a", "1"
    2) bent "{start}{start_dir}{bend-position}{new_dir}..." e.g.
        - c1+2<b+
    """

    b = Board(size)
    if len(threat) == 1:
        if threat in ascii_lowercase:
            return exec_road(threat + "1+", size)
        return exec_road("a" + threat + ">", size)
    srow, col, direction, *road = threat
    row, col = tile_to_coords(srow + col)

    sq = b.get(row, col)
    sqs = [coords_to_tile(sq.x, sq.y)]
    for bend, nd in zip(road[::2], road[1::2]):
        cond = True
        while cond:
            if direction in UP + DOWN:
                cond = sq.x < int(bend) - 1
            else:
                cond = sq.y < ascii_lowercase.index(bend)
            if not cond:
                direction = nd
            sq = b.get(*sq.next(direction, size))
            sqs.append(coords_to_tile(sq.x, sq.y))

    while True:
        nsq = b.get(*sq.next(direction, size))
        if (nsq.x, nsq.y) == (sq.x, sq.y):
            break
        sq = nsq
        sqs.append(coords_to_tile(sq.x, sq.y))

    if return_size:
        return sqs, size
    return sqs



from bredon.model import *
import random

class Action:
    def __init__(self, move, player):
        self.move = move
        try:
            self.color = player.color
        except:
            self.color = player

    def __repr__(self):
        return str(self.color) + ":" + str(self.move)

    @staticmethod
    def of(s):
        c, m = s.split(":")
        return Action(Move.of(m), Colors.of(c))

class Actions:
    def __init__(self, env):
        self.env = env

    def sample(self, player):
        return random.choice(list(self.env.get_actions(player)))


from .constants import *

from collections   import namedtuple
from dataclasses   import dataclass, field
from functools     import partial
from itertools     import combinations, chain, starmap, repeat
from operator      import sub
from tabulate      import tabulate, re
from string        import ascii_lowercase
from random        import randint

PseudoBoard = namedtuple("PseudoBoard",
                         ("w", "h", "board", "bool", "err", "type"))


def sums(n):
    b, mid, e = [0], list(range(1, n)), [n]
    splits = (d for i in range(n) for d in combinations(mid, i))
    return (list(map(sub, chain(s, e), chain(b, s))) for s in splits)


def tile_to_coords(t: str):
    return int(t[1]) - 1, ascii_lowercase.index(t[0])


def coords_to_tile(x: int, y: int):
    return ascii_lowercase[y] + str(x + 1)


@dataclass
class Move:
    total: int = 1
    stone: str = FLAT
    col: str = None
    row: int = None
    moves: list = field(default_factory=list)
    direction: str = None

    def get_square(self):
        return self.col + str(self.row)

    def __repr__(self):
        return 'Move(' + ', '.join([f"{k}={v!r}" for k, v in self.__dict__.items()]) + ')'

    def __str__(self):
        if not self.direction:
            return (self.stone + self.get_square()).strip(FLAT)
        else:
            return str(self.total) + self.get_square() + self.direction + ''.join(map(str, self.moves))

    @staticmethod
    def of(s):
        return str_to_move(s)

class Next:
    __slots__ = 'x', 'y'

    def next(self, direction, size):
        if direction == LEFT and self.y > 0:
            return self.x, self.y - 1
        if direction == RIGHT and self.y < size - 1:
            return self.x, self.y + 1
        if direction == DOWN and self.x > 0:
            return self.x - 1, self.y
        if direction == UP and self.x < size - 1:
            return self.x + 1, self.y
        return self.x, self.y

    def tru_next(self, direction, size):
        if direction == DOWN and self.y > 0:
            return self.x, self.y - 1
        if direction == UP and self.y < size - 1:
            return self.x, self.y + 1
        if direction == RIGHT and self.x > 0:
            return self.x - 1, self.y
        if direction == LEFT and self.x < size - 1:
            return self.x + 1, self.y
        return self.x, self.y


@dataclass
class Tile(Next):
    color: str
    stone: str = FLAT
    x: int = None
    y: int = None

    def __repr__(self):
        return '%s{%s}' % (self.color, self.stone)  # + f'@{coords_to_tile(self.x, self.y)}'


class Square(Next):
    def __init__(self, x, y, tiles=None):
        self.x, self.y = x, y
        self.tiles = [] if tiles is None else tiles

    def fix(self, tile, x=None, y=None):
        if x is None:
            tile.x, tile.y = self.x, self.y
        else:
            tile.x, tile.y, self.x, self.y = x, y, x, y

    def add(self, tile: Tile):
        self.tiles.append(tile)
        self.fix(tile)
        return self

    def extend(self, tiles):
        for tile in tiles:
            self.fix(tile)
        self.tiles.extend(tiles)
        return self

    def remove_top(self, n_tiles: int):
        n = len(self.tiles) - n_tiles
        top = self.tiles[n:]
        self.tiles = self.tiles[:n]
        return top

    def connections(self, board):
        conns = 0
        s = len(board)
        for direction in DIRS:
            x, y = self.next(direction, s)
            t_next = board[y][x].tiles
            if t_next:
                t_next = t_next[-1]
                t = self.tiles
                if t:
                    t = t[-1]
                    if t_next is not t and \
                            t_next.color == t.color and t_next.stone in 'FC' and t.stone in 'FC':
                        conns += 1
        return conns

    def copy(self):
        return Square(self.x, self.y, tiles=self.tiles[:])

    def generate_tps(self):
        if len(self.tiles) == 0:
            return 'x'
        tps = ''.join(str(tile.color.value + 1) for tile in self.tiles)
        ending = self.tiles[-1].stone
        if ending in CAP + STAND:
            tps += ending
        return tps

    def __eq__(self, other):
        if other == EMPTY:
            return not bool(self.tiles)
        if isinstance(other, Square):
            return self.tiles == other.tiles
        if isinstance(other, Tile) and self.tiles:
            return self.tiles[-1] == other
        return False

    def __repr__(self):
        return ''.join(str(t) for t in self.tiles)  + f'@{coords_to_tile(self.x, self.y)}'


class Board:
    def __init__(self, size: int, board=None):
        self.size = size
        self.board = [[Square(x, y) for x in range(size)]
                      for y in range(size)] if board is None else board
        self.stones, self.caps = SIZES[size]

    def place(self, move: Move, curr_player):
        """
        Attempt to place the tile.
        Returns a PseudoBoard object.

        :param move
        :param curr_player
        :return nt was it placed?, board state
        """
        tile = Tile(curr_player, stone=move.stone)
        x, y = tile_to_coords(move.get_square())
        new_board = self.copy_board()
        if self.board[y][x] == EMPTY and \
                tile.x is None and tile.y is None:
            new_board[y][x].add(tile)
            return PseudoBoard(self.size, self.size, new_board, True, None, "place")
        return PseudoBoard(self.size, self.size, new_board, False, "Tile cannot be placed there", None)

    def move_single(self, old_square, new_square, n_tiles: int, first=False) -> PseudoBoard:
        """
        Move n_tiles from old_square to
        new_square.
        """
        if not isinstance(old_square, Square):
            if isinstance(new_square, tuple):
                old_square = self.get(*old_square)
            elif isinstance(new_square, str):
                old_square = self.get(*tile_to_coords(old_square))
        else:
            old_square = self.get(old_square.x, old_square.y)

        if not isinstance(new_square, Square):
            if isinstance(new_square, tuple):
                new_square = self.get(*new_square)
            elif isinstance(new_square, str):
                try:
                    new_square = self.get(*old_square.next(new_square, self.size))
                except ValueError:
                    return PseudoBoard(self.size, self.size, self.board, False,
                                       f"{old_square.x}, {old_square.y} is out of bounds for {new_square}", None)
            else:
                raise TypeError("new_square must be Square, tuple, or str, got: %s" % new_square.__class__)
        else:
            new_square = self.get(new_square.x, new_square.y)

        new_square = new_square.copy()
        old_square = old_square.copy()
        new_board = self.copy_board()
        if n_tiles <= len(old_square.tiles) - int(not first):
            valid = False
            flatten = False
            if len(new_square.tiles) == 0:
                valid = True
            else:
                new_stone = new_square.tiles[-1].stone
                old_stone = old_square.tiles[-1].stone
                if old_stone == CAP:
                    if new_stone == FLAT:
                        valid = True
                    elif new_stone == STAND and n_tiles == 1:
                        valid = True
                        flatten = True
                elif new_stone == FLAT:
                    valid = True
            if valid:
                new_board[new_square.y][new_square.x] = new_square.copy() \
                    .extend(old_square.remove_top(n_tiles))
                if flatten:
                    new_square.tiles[-1].stone = FLAT

                new_board[old_square.y][old_square.x] = old_square.copy()
                return PseudoBoard(self.size, self.size, new_board, True, None, "move")
            return PseudoBoard(self.size, self.size, new_board, False,
                               f"Tile is not flat: stone == {new_square.tiles[-1].stone}", None)
        return PseudoBoard(self.size, self.size, new_board, False,
                           f"Too many tiles: {n_tiles} > {len(old_square.tiles) - int(not first)}", None)

    def move(self, move: Move):
        copy = self.copy()
        sq = copy.get(*tile_to_coords(move.get_square()))
        yield self._run(lambda: copy.move_single(sq, move.direction, move.total, first=True), copy)
        for n in range(1, len(move.moves)):
            sq = copy.get(*sq.next(move.direction, self.size))
            yield self._run(lambda: copy.move_single(sq, move.direction, sum(move.moves[n:]), first=False), copy)

    def parse_move(self, move: Move, curr_player):
        if move.direction is None:
            return self.place(move, curr_player)
        else:
            return self.move(move)

    def force(self, pbs):
        if isinstance(pbs, PseudoBoard):
            return self.force([pbs])
        else:
            new_board = self.copy_board()
            for pb in pbs:
                if pb.err is not None:
                    return pb.err
                else:
                    new_board = pb.board
            self.board = new_board

    def force_move(self, s, curr_player):
        return self.force(self.parse_move(s, curr_player))

    def force_str(self, s, curr_player):
        return self.force(self.parse_move(str_to_move(s), curr_player))

    def valid(self, pbs):
        new_board = self.copy()
        return new_board.force(pbs) is None

    def valid_move(self, move, color):
        new_board = self.copy()
        return new_board.force_move(move, color) is None

    def valid_str(self, move, color):
        return self.valid_move(str_to_move(move), color)

    def winner(self, players, t=False):
        r = self.road()
        if r:
            return r
        elif all(sq != EMPTY for row in self.board for sq in row) or \
                any(player.out_of_tiles() for player in players):
            return self.flat_win(t=t, f=True)
        return None

    def flat_win(self, t=False, f=False):
        w, b = self.count_flats(Colors.WHITE), self.count_flats(Colors.BLACK)
        if f or all(len(sq.tiles) > 0 for row in self.board for sq in row):
            return ("TIE" if t else False) if w == b else Colors.WHITE if w > b else Colors.BLACK
        return False

    def road(self, out=False):
        for board in (self.board, zip(*self.board)):
            for color in Colors:
                if self._road_check(color, board, out=out):
                    return color
        return False

    def get(self, x: int, y: int) -> Square:
        return self.board[y][x]

    def copy_board(self):
        return [[s.copy() for s in r] for r in self.board]

    def copy(self):
        return Board(self.size, board=self.copy_board())

    def __repr__(self):
        return tabulate(self.board,
                        tablefmt="plain",
                        headers=list(range(1, self.size + 1)),
                        showindex=list(ascii_lowercase[:self.size]))

    def evaluate(self, color):
        return self._evaluate(color) - self._evaluate(color.flip())

    def execute(self, move, color):
        self.force_move(move, color)

    @classmethod
    def from_moves(cls, moves, size, colors=repeat(Colors.WHITE), stones=repeat(FLAT), board=None):
        b = cls(size) if board is None else board.copy()
        for m, s, c in zip(moves, stones, colors):
            b.force_str(s + m, c)
        return b

    def set(self, old_state):
        self.board = old_state

    def _valid(self, move, color):
        return self.valid(self.parse_move(move, color))

    def generate_valid_moves(self, turn, color, caps):
        return filter(lambda move: self._valid(move, color), self.generate_all_moves(turn, color, caps))

    def generate_all_moves(self, turn, color, caps):
        for y in range(self.size):
            for x in range(self.size):
                c, r = coords_to_tile(x, y)
                tile = self.get(x, y)
                if tile == EMPTY:
                    for stone in FLAT + STAND:
                        yield Move(stone=stone, col=c, row=r)
                    if caps < self.caps and turn > 1:
                        yield Move(stone=CAP, col=c, row=r)
                else:
                    if tile.tiles[-1].color == color:
                        for direction in DIRS:
                            try:
                                x1, y1 = tile.next(direction, self.size)
                                if 0 <= x1 < self.size and 0 <= y1 < self.size:
                                    for i in range(1, min(len(tile.tiles) + 1, self.size + 1)):
                                        if i == 1 and len(tile.tiles) == 1:
                                            yield Move(col=c, row=r, direction=direction)
                                        else:
                                            for moves in sums(i):
                                                if len(moves) == 1 and moves[0] == i:
                                                    yield Move(total=i, col=c, row=r, direction=direction)
                                                else:
                                                    yield Move(total=i, col=c, row=r, moves=moves, direction=direction)
                            except ValueError:
                                pass

    def generate_tps(self, turn=1, move=1):
        tps = '/'.join(','.join(map(Square.generate_tps, row)) for row in self.board)
        for i in range(self.size, 1, -1):
            tps = re.sub("(x,){%d}x" % i, "x" + str(i + 1), tps)
            tps = re.sub("(x,){%d}" % i, "x" + str(i) + ",", tps)

        return '[TPS "' + tps + '" %d %d]' % (turn, move)

    # STATIC PRIVATE HELPER METHODS
    def _run(self, fn, copy):
        try:
            pb = fn()
            copy.force(pb)
            return pb
        except IndexError as e:
            return PseudoBoard(self.size, self.size, [], False, e, None)

    def _evaluate_sq(self, color, sq):
        e = 0
        if sq.tiles:
            t = sq.tiles[-1]
            if t.color == color:
                e += sum(1 for i in sq.tiles if i.color == color and i.stone in 'CF') ** 5
                e += (sq.connections(self.board) + 1) ** 1.5
        return e

    def _evaluate(self, color):
        return sum(sum(map(partial(self._evaluate_sq, color), row)) for row in self.board) + \
               sum(map(len, self._compress_left(color, self.board, False))) ** 2 + \
               sum(map(len, self._compress_left(color, zip(*self.board), False))) ** 2

    def _cl_sq_check(self, r, color, out, sq):
        if out:
            print(r, color, sq)
        if sq.tiles and sq.tiles[-1].color == color:
            conns = sq.connections(self.board)
            return conns > 1 or ((r == 0 or r == self.size - 1) and conns > 0)
        return False

    def _cl_row_check(self, color, out, r, row):
        return list(filter(partial(self._cl_sq_check, r, color, out), row))

    def count_flats(self, color):
        return sum(1 for row in self.board for sq in row if sq.tiles and
                   sq.tiles[-1].color == color and sq.tiles[-1].stone == FLAT)

    def _compress_left(self, color, board, out):
        return list(starmap(partial(self._cl_row_check, color, out), enumerate(board)))

    def _road_check(self, color, board, out):
        road = self._compress_left(color, board, out)
        # if out:
        #     print(road)
        if road and (all(road) or
                     any(len(road[i]) >= self.size
                         for i in range(self.size))):
            return color
        return False

def str_to_move(move: str) -> Move:
    move_dir = None
    for direction in DIRS:
        if direction in move:
            move_dir = direction
            move = move.split(direction)
            break

    if move_dir is None:
        stone, c, r = move.zfill(3)
        return Move(stone=FLAT if stone == '0' else stone, col=c, row=r)
    else:
        ns = move[1]
        t = move[0]
        if t[0] not in ascii_lowercase:
            total = int(t[0])
            t = t[1:]
        else:
            total = 1
        c, r = t
        return Move(total=total, col=c, row=r, direction=move_dir, moves=list(map(int, ns)))



from .utils import *


class PTN:
    def __init__(self, turn=1):
        self.moves = []
        self.turn = turn

    def append(self, move):
        if (not self.moves) or len(self.moves[-1]) == 2:
            self.moves.append([move])
        else:
            self.moves[-1].append(move)

    def fill(self):
        if len(self.moves) > 1:
            return self.moves
        m = [move for ply in self.moves for move in ply]
        while len(m) < 2:
            m.append("")
        return [m]

    def get(self, turn, color: Colors):
        return self[turn][color.value]

    def clear(self):
        self.moves.clear()

    def __getitem__(self, turn):
        return self.moves[turn]

    def __repr__(self):
        return "PTN(%r)" % self.moves

    def __str__(self):
        return tabulate(self.fill(), tablefmt="plain",
                        headers=[i.name for i in Colors],
                        showindex=range(self.turn, len(self.fill())+self.turn))

    def __len__(self):
        return len(self.moves)

    def __iter__(self):
        if self.moves:
            return iter(self.moves)
        return iter([])


def parse_tps(tps):
    rows = [i.split(",") for i in tps.strip('[TPS"] ').split("/")]
    last, move, turn = rows[-1][-1].split()
    rows[-1][-1] = last
    fs, cs = [0, 0], [0, 0]
    for y, row in enumerate(rows):
        r = []
        for x, c in enumerate(row):
            if 'x' == c:
                r.append(Square(x, y))
            elif 'x' in c:
                r.extend([Square(x, y) for _ in range(int(c[1]))])
            else:
                tiles = []
                for stone in c[:-1]:
                    tiles.append(Tile(Colors(int(stone)-1), x=x, y=y))
                    fs[int(stone)-1] += 1
                if c[-1] in STONES:
                    tiles[-1].stone = c[-1]
                    if c[-1] == CAP:
                        cs[int(stone)-1] += 1
                else:
                    tiles.append(Tile(Colors(int(c[-1])-1), x=x, y=y))
                r.append(Square(x, y, tiles=tiles))
        rows[y] = r[:]
    board = list(zip(*reversed(rows)))
    for y, row in enumerate(board):
        for x, sq in enumerate(row):
            sq.x, sq.y = x, y
            for tile in sq.tiles:
                sq.fix(tile=tile)
    return Board(len(rows[0]), board=board), move, turn, [fs[0], cs[0]], [fs[1], cs[1]]



from model import *

TEST_SIZE = 4


def test_straight_roads():
    def _test(c, z2):
        b = Board(TEST_SIZE)
        ms = ""
        for m in zip(c if len(c) == TEST_SIZE else c * TEST_SIZE, z2):
            b.force_str(''.join(m), Colors.WHITE)
            ms += ''.join(m) + " "
            if len(ms.split()) < TEST_SIZE:
                try:
                    assert b.road() == False
                    print(ms, "worked!")
                except:
                    print(ms, "did not work")
                    print(b)
                    print(b.road())
                    input()
        try:
            assert b.road() == Colors.WHITE
            print(ms, "worked!")
        except:
            print(ms, "did not work")
            print(b)
            print(b.road())
            input()

    for c in "abcd":
        _test(c, "1234")
    for d in "1234":
         _test("abcd", d * TEST_SIZE)

test_straight_roads()


from .view import *

import click

def validate_type(ctx, param, value):
    t, *n = value.split(":")
    t = t.upper()
    if t[0] == "H" and t in "HUMAN":
        return HUMAN
    elif t[0] == "A" and t in "AI":
        try:
            return AI(int(n[0]))
        except ValueError:
            raise click.BadParameter("depth must be a valid integer")
    else:
        raise click.BadParameter("type must be either 'human' or 'ai'")


def validate_game(ctx, param, value):
    if not value.endswith("Game"):
        value += "Game"
    try:
        return globals()[value]
    except KeyError:
        raise click.BadParameter("must be TextGame or ViewGame")


def validate_size(ctx, param, value):
    if 3 <= value <= 8:
        return value
    raise click.BadParameter("size must be between 3 and 8")


@click.command()
@click.option("--game",  "-g", callback=validate_game, default="ViewGame",
              help="Type of game to play: [Text|View]Game", show_default=True)
@click.option("--size",  "-s", callback=validate_size, default=5,
              help="The size of the board", show_default=True)
@click.option("--white", "-w", callback=validate_type, default="human",
              help="Type of the white player, [h(uman)|a(i)]:depth", show_default=True)
@click.option("--black", "-b", callback=validate_type, default="ai:3",
              help="Type of the black player, [h(uman)|a(i)]:depth", show_default=True)
def run_game(game, size, white, black):
    game(size=size, white=white, black=black).run()
    test_notation()
    # unittest.main(exit=False)


def test_notation():
    print(parse_tps('[TPS "x3,12,2S/x,22S,22C,11,21/121,212,12,1121C,1212S/21S,1,21,211S,12S/x,21S,2,x2 1 26"]'))
    *moves, (b, ptn) = load_moves_from_file("test-game.ptn")
    print(b)
    print(str(ptn))


if __name__ == '__main__':
    run_game()


import numpy as np
import gym

from keras import *

def keras_learning(env, num_episodes=500):
    model = Sequential()
    model.add(InputLayer(batch_input_shape=(1, 5)))
    model.add(Dense(10, activation='sigmoid'))
    model.add(Dense(2, activation='linear'))
    model.compile(loss='mse', optimizer='adam', metrics=['mae'])

    y = 0.95
    eps = 0.5
    decay_factor = 0.999
    r_avg_list = []
    for i in range(num_episodes):
        s = env.reset()
        eps *= decay_factor
        if i % 100 == 0:
            print("Episode {} of {}".format(i + 1, num_episodes))
        done = False
        r_sum = 0
        while not done:
            if np.random.random() < eps:
                a = np.random.randint(0, 2)
            else:
                a = np.argmax(model.predict(np.identity(5)[s:s + 1]))
            new_s, r, done, _ = env.step(a)
            target = r + y * np.max(model.predict(np.identity(5)[new_s:new_s + 1]))
            target_vec = model.predict(np.identity(5)[s:s + 1])[0]
            target_vec[a] = target
            model.fit(np.identity(5)[s:s + 1], target_vec.reshape(-1, 2), epochs=1, verbose=0)
            s = new_s
            r_sum += r
        r_avg_list.append(r_sum / 1000)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="bredon",
    version="1.0.4",
    author="Simon Chervenak",
    author_email="simonlcherv@gmail.com",
    description="A set of tools for the game Tak.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Innoviox/bredon",
    packages=setuptools.find_packages(),
    license="GNU General Public License v3.0",
    install_requires=['dataclasses==0.4',
                      'tabulate==0.8.2',
                      'click==7.0',
                      'gym==0.10.8',
                      'tqdm==4.26.0']
)


import random

from .utils import *

inf = float('inf')


class Player(object):
    def __init__(self, board, color, name=None):
        self.board = board
        self.color = color
        self.stones, self.caps = 0, 0
        self.name = self.color.name if name is None else name

    def _do(self, m: Move, c, f=True):
        move = self.board.parse_move(m, c)
        if isinstance(move, PseudoBoard):
            if move.bool:
                stone_type = m.stone
                stone, cap = False, False
                if stone_type in [FLAT, STAND]:
                    self.stones += 1
                    stone = True
                else:
                    self.caps += 1
                    cap = True
                caps, _stones = self.caps > self.board.caps, self.stones > self.board.stones
                if _stones and caps:
                    self.stones -= stone
                    self.caps -= cap
                    # TODO: End the game
                    self.board.end_game()
                elif (_stones and stone) or (caps and cap):
                    self.stones -= stone
                    self.caps -= cap
                    raise ValueError(f"Not enough pieces left")
                    # f"\tStones played: {self.stones}, Total: {self.board.stones}"
                    # f"\tCaps played: {self.caps}, Total: {self.board.caps}"
                    # f"Stone: {stone}, Cap: {cap}")
                elif f:
                    return self.board.force(move)
            else:
                # Move is illegal
                raise ValueError("Illegal Move")
        elif f:
            self.board.force(move)

    def do(self, m: Move):
        return self._do(m, self.color)

    def out_of_tiles(self):
        return self.caps >= self.board.caps and self.stones >= self.board.stones

    def _pick_move(self, turn, color, input_fn=input):
        while True:
            m = str_to_move(input_fn("Enter move: "))
            # try:
            v = self.board.copy().force_move(m, color)
            if v is None:
                return m, color
            else:
                print("Parsed move", m)
                input("Received error "+ v)
            # except Exception as e:
            #     print("Parsed move", m)
            #     print("Received error", e)

    def pick_move(self, turn, input_fn=input, out=False):
        return self._pick_move(turn, self.color, input_fn=input_fn)[0]

    def pick_opposing_move(self, turn, input_fn=input, out=False):
        return self._pick_move(turn, self.color.flip(), input_fn=input_fn)


class BaseAI(Player):
    def pick_move(self, turn, input_fn=input, out=False):
        return

class RandomAI(BaseAI):
    def pick_move(self, turn, input_fn=input, out=False):
        return random.choice(list(self.board.generate_valid_moves(self)))


class MinimaxAI(BaseAI):
    def __init__(self, board, color, depth=3):
        super().__init__(board, color)
        self.depth = depth

    def pick_opposing_move(self, turn, input_fn=input, out=False):
        # if self.board.valid_str("a1", self.color.flip()):
        #     return str_to_move("a1"), self.color.flip()
        # return str_to_move(ascii_lowercase[self.board.size - 1] + str(self.board.size)), self.color.flip()
        return self.pick_move(turn, input_fn=input_fn), self.color.flip()

    def pick_move(self, turn, input_fn=None, out=False) -> Move:
        moves = self.board.generate_valid_moves(turn, self.color, self.caps)
        best_eval = inf
        if self.depth % 2 == 1:
            best_eval *= -1
        best_move = None
        old_state = self.board.copy_board()
        alpha = -inf
        for move in moves:
            if out: print("Running move", move, "current alpha", alpha)
            self.board.execute(move, self.color)
            alpha = self.minimax(self.depth - 1, self.board, alpha, inf, True, self.color.flip(),
                                 self.board.copy_board(), turn, out=out) * 4
            ev = self.board.evaluate(self.color)
            alpha -= ev / 2
            if abs(alpha) > THRESHOLD:
                alpha *= -1
            self.board.set(old_state)
            if out: print("\t", ev, alpha, best_eval, best_move)
            if (self.depth % 2 == 1 and alpha >= best_eval) or (self.depth % 2 == 0 and alpha <= best_eval):
                best_eval = alpha
                best_move = move
                if out: print("\tsetting move")
        if out: input()
        return best_move

    def minimax(self, depth, board, alpha, beta, maximising, color, old_state, turn, out=False):
        if board.road() or board.flat_win():
            if maximising:
                return -MAX_N * depth
            else:
                return MAX_N * depth
        elif depth == 0:
            return board.evaluate(color)
        
        moves = board.generate_valid_moves(turn, color, self.caps)
        if maximising:
            b_eval = -inf
            for move in moves:
                board.execute(move, color)
                b_eval = max(b_eval, self.minimax(depth - 1, board, alpha, beta, not maximising, color.flip(),
                                                  board.copy_board(), turn, out=out))
                board.set(old_state)
                alpha = max(alpha, b_eval)
                if beta <= alpha:
                    if beta < -THRESHOLD:
                        return beta
                    return -beta
        else:
            b_eval = inf
            for move in moves:
                board.execute(move, color)
                b_eval = min(b_eval, self.minimax(depth - 1, board, alpha, beta, not maximising, color.flip(),
                                                  board.copy_board(), turn, out=out))
                board.set(old_state)
                beta = min(beta, b_eval)
                if beta <= alpha:
                    if beta < -THRESHOLD:
                        return beta
                    return -beta
        return b_eval



import os
from model import load_moves_from_file

if not os.path.exists("games-csv/"):
    os.makedirs("games-csv/")

def convert(infile):
    with open(infile) as f:
        with open("games-csv/" + infile.split("/")[1][:-4] + ".csv", "w") as o:
            ptn = list(load_moves_from_file(infile))[-1][-1]
            prev = ''
            for move in ptn:
                for a in move:
                    s = str(a)
                    if prev:
                        o.write(prev+","+s+"\n")
                    prev += s


for a, b, c in os.walk("games/"):
    for k in c:
        convert(a+k)
        # input()


import gym
from bredon.model import *
from .actions import Action
import random
from bredon.view import *

class TakEnv(gym.Env):
    metadata = {'render.modes': ['ansi', 'human']}

    def __init__(self, size=5, players=None, board=None):
        self.size = size
        self.board = Board(size=size, board=board)
        self.players = players if players else []
        self.turn, self.move = 1, 1
        self.done = False
        self.vg = None

    def get_actions(self, player):
        return map(lambda i: Action(i, player), self.board.generate_valid_moves(self.turn, player.color, player.caps))

    def _step_turn(self):
        self.move = self.move % 2 + 1
        if self.move == 1:
            self.turn += 1

    def step(self, action):
        move = self.board.parse_move(action.move, action.color)
        if isinstance(move, PseudoBoard) and move.bool:
            self.board.force(move)
        else:
            for m in move:
                if isinstance(move, PseudoBoard) and move.bool:
                    self.board.force(m)
                else:
                    return False

        self._step_turn()
        s, r, d, i = (self.board.board,
                      self.board.evaluate(action.color),
                      self.board.winner(self.players),
                      {} # TODO: implement info
                     )
        if d:
            self.done = True
        return s, r, d, i

    def reset(self):
        self.board = Board(size=self.size)

    def render(self, mode='ansi'):
        if mode == 'ansi':
            print(str(self.board))
        elif mode == 'human':
            if not self.vg:
                self.vg = ViewGame(size=self.board.size, board=self.board, white=HUMAN, black=HUMAN)
            else:
                self.vg._init_gui(board=self.board)
                self.vg.viz()

    def close(self):
        self.board = None

    def seed(self, seed=None):
        random.seed(seed)
        return [seed]

    def add_player(self, p: Player):
        self.players.append(p)

class HypoEnv(TakEnv):
    def __init__(self, **kwargs):
        TakEnv.__init__(self, **kwargs)
        self.old_state = self.board.copy()
        
    def step(self, action):
        ret = super().step(action)
        self.old_state = self.board.copy()
        return ret
        
    def hypostep(self, action):
        return super().step(action)
    
    def unstep(self):
        self.board = self.old_state.copy()

def hypoenv(env: TakEnv):
    return HypoEnv(size=env.size, players=env.players, board=env.board.board)
